/*  ==================================================================	*
 *				Editor mined				*
 *				Search functions			*
 *  ==================================================================	*/

#include "mined.h"

/* external */
extern FLAG redraw_pending;	/* was a redraw suppressed in find_y ? */

/* forward */
void search ();
void re_search ();
void prev_search ();
int line_check ();
int check_string ();
int in_exprlist ();
int star ();
void search_corresponding ();
LINE * search_it ();
REGEX * make_expression ();

/* local */
FLAG wrapped_search;		/* temporary flag of wrapped search */


/*  ==================================================================	*
 *				Search Commands				*
 *  ==================================================================	*/

/*
 * A regular expression consists of a sequence of:
 *	1. A normal character matching that character.
 *	2. A . matching any character.
 *	3. A ^ matching the begin of a line.
 *	4. A $ (as last character of the pattern) mathing the end of a line.
 *	5. A \<character> matching <character>.
 *	6. A number of characters enclosed in [] pairs matching any of these
 *	   characters. A list of characters can be indicated by a '-'. So
 *	   [a-z] matches any letter of the alphabet. If the first character
 *	   after the '[' is a '^' then the set is negated (matching none of
 *	   the characters).
 *	   A ']', '^' or '-' can be escaped by putting a '\' in front of it.
 *	7. If one of the expressions as described in 1-6 is followed by a
 *	   '*' than that expressions matches a sequence of 0 or more of
 *	   that expression.
 */

char typed_expression [maxLINE_LEN];	/* Holds previous search expression */

/*
 * SIDFW searches forward for the current identifier.
 */
void
SIDFW ()
{
  SIDF (FORWARD);
}

/*
 * SIDRV searches backward for the current identifier.
 */
void
SIDRV ()
{
  SIDF (REVERSE);
}

/*
 * SFW searches forward for an expression.
 */
void
SFW ()
{
  if (hop_flag > 0) SIDFW ();
  else search ("Search forward:", FORWARD);
}

/*
 * SRV searches backward for an expression.
 */
void
SRV ()
{
  if (hop_flag > 0) SIDRV ();
  else search ("Search reverse:", REVERSE);
}

/*
 * RS searches using the last search direction and expression.
 */
void
RS ()
{
  if (hop_flag > 0) prev_search ();
  else re_search ();
}

/*
 * Get_expression () prompts for an expression. If just a return is typed, the
 * old expression is used. If the expression changed, compile () is called and
 * the returning REGEX structure is returned. It returns NIL_REG upon error.
 * The save flag indicates whether the expression should be appended at the
 * message pointer.
 */
char exp_buf [maxLINE_LEN];		/* Buffer for new expr. */

REGEX *
get_expression (message)
  char * message;
{
  static REGEX program;			/* Program of expression */

  if (get_string (message, exp_buf, False, "") == ERRORS)
	return NIL_REG;

  if (exp_buf [0] == '\0' && typed_expression [0] == '\0') {
	error ("No previous search expression", NIL_PTR);
	return NIL_REG;
  }

  if (exp_buf [0] != '\0') {			/* A new expr. is typed */
	copy_string (typed_expression, exp_buf);	/* Save expr. */
	/* Compile new expression: */
	if (compile (exp_buf, & program) == ERRORS)
		return NIL_REG;
  }

  if (program.status == REG_ERROR) {	/* Error during compiling */
	error (program.result.err_mess, NIL_PTR);
	return NIL_REG;
  }
  return & program;
}

/*
 * make_expression is only called by search_for and maintains its own, 
 * independant search program buffer
 */
REGEX *
make_expression (expr)
  char * expr;
{
  static REGEX program;			/* Program of expression */

  if (compile (expr, & program) == ERRORS)	/* Compile new expression */
	return NIL_REG;

  if (program.status == REG_ERROR) {	/* Error during compiling */
	error (program.result.err_mess, NIL_PTR);
	return NIL_REG;
  }
  return & program;
}

/*
 * Change () prompts for an expression and a substitution pattern and changes
 * all matches of the expression into the substitution.
 * change () starts looking for expressions at the current line and
 * continues until the end of the file if the FLAG `global' is VALID.
 * It prompts for each change if the FLAG `confirm' is True.
 * For a call graph of search procedures see search ().
 */
void
change (message, global, confirm)
  char * message;		/* Message to prompt for expression */
  FLAG global, confirm;
{
  char mess_buf [maxLINE_LEN];		/* Buffer to hold message */
  char replacement [maxLINE_LEN];	/* Buffer to hold subst. pattern */
  REGEX * program;			/* Program resulting from compilation */
  register LINE * line = cur_line;
  char * textp;
  char * substitute ();
  long lines = 0L;		/* Nr of lines on which subs occurred */
  long subs = 0L;		/* Nr of subs made */
  int ly = y;			/* Index to check if line is on screen */
  int previousy = y;
  char c;
  FLAG quit_change;

  if (viewonly == True)
	{viewonlyerr (); return;}

/* Save message and get expression */
  if ((program = get_expression (message)) == NIL_REG)
	return;

/* Get substitution pattern */
  build_string (mess_buf, "%s %s by:", message, typed_expression);
  if (get_string (mess_buf, replacement, False, "") == ERRORS)
	return;

  set_cursor (0, YMAX);
  flush ();
/* Substitute until end of file */
  do {
	if (line_check (program, line->text, FORWARD)) {
	    lines ++;
	    /* Repeat sub. on this line as long as we find a match */
	    do {
		if (confirm == True) {
			ly = find_y (line);
			textp = program->start_ptr;
			move_address (program->start_ptr, ly);
			status_msg ("Replace ? (y/n)");
			c = promptyn ();
			clear_status ();
			if (c == 'y') {
			    subs ++;	/* Increment subs */
			    if ((textp = substitute (line, program, replacement))
							== NIL_PTR)
				    return;	/* Line too long */
			    set_cursor (0, ly);
			    line_print (line);
			}
			else
			    advance_char (& textp);
		}
		else {
		    subs ++;	/* Increment subs */
		    line->shift_count = 0;
			/* in case line would get completely shifted out */
		    if ((textp = substitute (line, program, replacement))
							== NIL_PTR) {
			set_cursor (0, ly);
			line_print (line);
			move_to (x, y);
			return;	/* Line too long */
		    }
		}
	    } while (  (program->status & BEGIN_LINE) != BEGIN_LINE
		    && (program->status & END_LINE) != END_LINE
		    && line_check (program, textp, FORWARD)
		    && quit == False);
	    /* Check to see if we can print the result */
	    if (confirm == False && ly <= SCREENMAX) {
		set_cursor (0, ly);
		line_print (line);
	    }
	}
	if (ly <= SCREENMAX)
		ly ++;
	line = line->next;
  } while (line != tail && global == VALID && quit == False);

  quit_change = quit;
/* Fix the status line */
  if (subs == 0L && quit == False)
	error ("Pattern not found", NIL_PTR);
  else if (lines >= REPORT || quit == True) {
	build_string (mess_buf, "%s %ld substitutions on %ld lines",
		(quit_change == True) ? "(Aborted) " : "", subs, lines);
	status_msg (mess_buf);
  }
  else if (global == NOT_VALID && subs >= REPORT)
	status_line (num_out (subs), " substitutions");
  else
	clear_status ();

  if (confirm == True) move_to (x, y);
  else move_to (LINE_START, previousy);

/*  if (quit == True) swallow_dummy_quit_char (); */
  quit = False;
}

/*
 * Substitute () replaces the match on this line by the substitute pattern
 * as indicated by the program. Every '&' in the replacement is replaced by
 * the original match. A \ in the replacement escapes the next character.
 */
char *
substitute (line, program, replacement)
  LINE * line;
  REGEX * program;
  char * replacement;	/* Contains replacement pattern */
{
  register char * textp = text_buffer;
  register char * subp = replacement;
  char * linep = line->text;
  char * amp;
  char * newtext;

  set_modified ();

/* Copy part of line until the beginning of the match */
  while (linep != program->start_ptr)
	* textp ++ = * linep ++;

/*
 * Replace the match by the substitution pattern. Each occurrence of '&' is
 * replaced by the original match. A \ escapes the next character.
 */
  while (* subp != '\0' && textp < & text_buffer [MAX_CHARS]) {
	if (* subp == '&') {		/* Replace the original match */
		amp = program->start_ptr;
		while (amp < program->end_ptr && textp < & text_buffer [MAX_CHARS])
			* textp ++ = * amp ++;
		subp ++;
	}
	else {
		if (* subp == '\\' && * (subp + 1) != '\0')
			subp ++;
		* textp ++ = * subp ++;
	}
  }
  * textp = '\0';

/* Check for line length not exceeding MAX_CHARS */
  if (length_of (text_buffer) + length_of (program->end_ptr) >= MAX_CHARS) {
	error ("Substitution failed: resulted line too long", NIL_PTR);
	return NIL_PTR;
  }

/* Append last part of line to the newly built line */
  copy_string (textp, program->end_ptr);

/* Free old line and install new one */
  newtext = alloc (length_of (text_buffer) + 1);
  if (newtext == NIL_PTR) {
	ring_bell ();
	error ("Substitution failed: cannot allocate more memory", NIL_PTR);
	return NIL_PTR;
  }
  else {
	free_space (line->text);
	line->text = newtext;
	copy_string (line->text, text_buffer);
	return (line->text + (int) (textp - text_buffer));
  }
}

/*
 * GR () a replaces all matches from the current position until the end
 * of the file.
 */
void
GR ()
{
  change ("Global replace:", VALID, False);
}

/*
 * Replace is substitute with confirmation dialogue.
 */
void
REPL ()
{
  change ("Global replace (with confirm):", VALID, True);
}

/*
 * LR () replaces all matches on the current line.
 */
void
LR ()
{
  change ("Line replace:", NOT_VALID, False);
}

/*
 * Search () calls get_expression to fetch the expression. If this went well,
 * the function match () is called which returns the line with the next match.
 * If this line is the NIL_LINE, it means that a match could not be found.
 * Find_x () and find_y () display the right page on the screen, and return
 * the right coordinates for x and y.
 * These coordinates are passed to move_to ().
 * Re_search () searches using the last search program and direction.
   Call graph of search procedures:
RS ------------\
SFW -\		> re_search > match -----------\
SRV --> search <				\
	        \				 \
	         > get_expression > compile	  > line_check > check_string
GR  \	        /				 /
REPL > change  <--------------------------------/
LR  /	        \
		 \ substitute

 */
REGEX * lastprogram = NIL_REG;
FLAG lastmethod = NOT_VALID;	/* FORWARD, REVERSE */

char prevexpr [maxLINE_LEN];	/* Buffer for previous expr. */
FLAG prevmethod = NOT_VALID;	/* FORWARD, REVERSE */

void do_search ();

void
re_search ()
{
  do_search (lastprogram, lastmethod);
}

void
prev_search ()
{
  search_for (prevexpr, prevmethod);
}

void
search (message, method)
  char * message;
  FLAG method;
{
  register REGEX * program;
  char prevexp_buf [maxLINE_LEN];	/* Buffer for previous expr. */

  if (lastmethod != NOT_VALID) copy_string (prevexp_buf, exp_buf);

/* Get the expression */
  if ((program = get_expression (message)) == NIL_REG)
	return;

  if (program != NIL_REG && lastmethod != NOT_VALID) {
	copy_string (prevexpr, prevexp_buf);
	prevmethod = lastmethod;
  }
  lastprogram = program;
  lastmethod = method;

  re_search ();
}

void
search_for (expr, method)
  char * expr;
  FLAG method;
{
  register REGEX * program;

/* make the expression */
  if ((program = make_expression (expr)) == NIL_REG)
	return;
  do_search (program, method);
}

/*
 * search_corresponding searches for a corresponding bracket of a 
 * bracket pair expression.
 * The parameter `method` must be valid.
 */
void
search_corresponding (expr, method, corresponding)
  char * expr;
  FLAG method;
  char corresponding;
{
  register REGEX * program;
  LINE * match_line;
  int bracount = 0;
  LINE * old_line = cur_line;
  char * old_text = cur_text;
  FLAG was_redraw_pending;
/*  char pflags [99]; */

/* make the expression */
  if ((program = make_expression (expr)) == NIL_REG)
	return;

  redraw_pending = False;
/* strcpy (pflags, redraw_pending == True ? "T" : "F"); */
  wrapped_search = False;
  do {
	match_line = search_it (program, method);
	if (match_line == NIL_LINE)
		return;
	if (wrapped_search == True) {
		status_msg ("No corresponding bracket");
		move_address (old_text, find_y (old_line));
		return;
	}
	if (* (program->start_ptr) == corresponding)
		bracount --;
	else	bracount ++;
	if (bracount >= 0) {
	/* set position to match for next search: */
		move_address_w_o_RD (program->start_ptr, find_y_w_o_RD (match_line));
/* strcat (pflags, redraw_pending == True ? "T" : "F"); */
	}
/*
	build_string (text_buffer, "bracount = %d", bracount);
	status_msg (text_buffer);
	sleep (1);
	break;
*/
	if (quit == True) {
		status_msg ("Breacket search aborted");
		quit = False;
		move_address (old_text, find_y (old_line));
		return;
	}
  } while (bracount >= 0);

/* set position to old to enable correct function of find_y: */
  move_address_w_o_RD (old_text, find_y_w_o_RD (old_line));
/* strcat (pflags, redraw_pending == True ? "T" : "F"); */

/* position and re-display if necessary: */
  was_redraw_pending = redraw_pending;
  move_address (program->start_ptr, find_y (match_line));
/* strcat (pflags, redraw_pending == True ? "T" : "F"); */
  if (was_redraw_pending == True && redraw_pending == True) {
	RD ();
/*	status_line ("pending redraw performed ", pflags);	*/
  }
}

/*
 * do_search checks its parameters, does the search, and positions
 * For search_it, arguments must be valid, and it does not position
 */
LINE *
search_it (program, method)
  register REGEX * program;
  FLAG method;
{
  register LINE * match_line;

  set_cursor (0, YMAX);
  clear_status ();
  flush ();
/* Find the match */
  if ((match_line = match (program, cur_text, method)) == NIL_LINE) {
	if (quit == True) {
		status_msg ("Aborted");
/*		swallow_dummy_quit_char (); */
		quit = False;
	}
	else
		status_msg ("Pattern not found");
  }
  return match_line;
}

void
do_search (program, method)
  register REGEX * program;
  FLAG method;
{
  register LINE * match_line;

  if (method == NOT_VALID) {
	error ("No previous search", NIL_PTR);
	return;
  }
  if (program == NIL_REG) {
	error ("No previous search expression", NIL_PTR);
	return;
  }

  match_line = search_it (program, method);
  if (match_line == NIL_LINE) return;
  move_address (program->start_ptr, find_y (match_line));
}


/* Now the search and replace utilities */
/* ------------------------------------ */
/* Opcodes for characters */
#define	NORMAL		0x0200
#define DOT		0x0400
#define EOLN		0x0800
#define STAR		0x1000
#define BRACKET		0x2000
#define NEGATE		0x0100
#define DONE		0x4000

/* Mask for opcodes and characters */
#define LOW_BYTE	0x00FF
#define HIGH_BYTE	0xFF00

/* Previous is the contents of the previous address (ptr) points to */
#define previous(ptr)		(* ((ptr) - 1))

/* Buffer to store outcome of compilation */
int exp_buffer [BLOCK_SIZE];

/* Errors often used */
char * too_long = "Regular expression too long";

/*
 * Reg_error () is called by compile () if something went wrong. It sets the
 * status of the structure to error, and assigns the error field of the union.
 */
#define reg_error(str)	program->status = REG_ERROR, \
					program->result.err_mess = (str)

/*
 * Bcopy copies `bytes' bytes from the `from' address into the `to' address.
 */
#ifdef vms
#define defbcopy
#endif
#ifdef msdos
#define defbcopy
#endif

#ifdef defbcopy	/* otherwise also in standard library */
void
bcopy (from, to, bytes)
  register char * from, * to;
  register int bytes;
{
  while (bytes --)
	* to ++ = * from ++;
}
#else
#ifdef sysV
#define bcopy(from, to, len)	memcpy (to, from, len)
#else
extern void bcopy ();
#endif
#endif

/*
 * Finished () is called when everything went right during compilation. It
 * allocates space for the expression, and copies the expression buffer into
 * this field.
 */
int
finished (program, last_exp)
  register REGEX * program;
  int * last_exp;
{
  register unsigned int length = (last_exp - exp_buffer) * sizeof (int);
/* Allocate space */
  program->result.expression = (int *) alloc (length);
  if (program->result.expression == NIL_INT) {
	ring_bell ();
	error ("Cannot allocate memory for search expression", NIL_PTR);
	return ERRORS;
  }
  else {	/* Copy expression. (expression consists of ints!) */
	bcopy (exp_buffer, program->result.expression, length);
	return FINE;
  }
}

/*
 * Compile compiles the pattern into a more comprehensible form and returns a
 * REGEX structure. If something went wrong, the status field of the structure
 * is set to REG_ERROR and an error message is set into the err_mess field of
 * the union. If all went well the expression is saved and the expression
 * pointer is set to the saved (and compiled) expression.
 */
int
compile (pattern, program)
  register character * pattern;	/* Pointer to pattern */
  REGEX * program;
{
  register int * expression = exp_buffer;
  int * prev_char;		/* Pointer to previous compiled atom */
  int * acct_field = NIL_INT;	/* Pointer to last BRACKET start */
  FLAG negate;			/* Negate flag for BRACKET */
  character low_char;		/* Index for chars in BRACKET */
  character c;

/* Check for begin of line */
  if (* pattern == '^') {
	program->status = BEGIN_LINE;
	pattern ++;
  }
  else {
	program->status = 0;
/* If the first character is a '*' we have to assign it here. */
	if (* pattern == '*') {
		* expression ++ = '*' + NORMAL;
		pattern ++;
	}
  }

  for (; ;) {	c = * pattern ++;
	switch (c) {
	case '.' :
		* expression ++ = DOT;
		break;
	case '$' :
		/*
		 * Only means EOLN if it is the last char of the pattern
		 */
		if (* pattern == '\0') {
			* expression ++ = EOLN | DONE;
			program->status |= END_LINE;
			return finished (program, expression);
		}
		else
			* expression ++ = NORMAL + '$';
		break;
	case '\0' :
		* expression ++ = DONE;
		return finished (program, expression);
	case '\\' :
		/* If last char, it must! mean a normal '\' */
		if (* pattern == '\0')
			* expression ++ = NORMAL + '\\';
		else
			* expression ++ = NORMAL + * pattern ++;
		break;
	case '*' :
		/*
		 * If the previous expression was a [] find out the
		 * begin of the list, and adjust the opcode.
		 */
		prev_char = expression - 1;
		if (* prev_char & BRACKET)
			* (expression - (* acct_field & LOW_BYTE)) |= STAR;
		else
			* prev_char |= STAR;
		break;
	case '[' :
		/*
		 * First field in expression gives information about
		 * the list.
		 * The opcode consists of BRACKET and if necessary
		 * NEGATE to indicate that the list should be negated
		 * and/or STAR to indicate a number of sequence of this
		 * list.
		 * The lower byte contains the length of the list.
		 */
		acct_field = expression ++;
		if (* pattern == '^') {	/* List must be negated */
			pattern ++;
			negate = True;
		}
		else
			negate = False;
		while (* pattern != ']') {
			if (* pattern == '\0') {
				reg_error ("Missing ]");
				return FINE;
			}
			if (* pattern == '\\')
				pattern ++;
			* expression ++ = * pattern ++;
			if (* pattern == '-') {
						/* Make list of chars */
				low_char = previous (pattern);
				pattern ++;	/* Skip '-' */
				if (utf8_text == True
				    && ((low_char & 0x80 == 0x80)
				     || (* pattern & 0x80 == 0x80))) {
					reg_error ("Range [..-..] not allowed with multi-byte characters");
					return FINE;
				} else
				if (low_char ++ > * pattern) {
					reg_error ("Bad range in [a-z]");
					return FINE;
				}
				/* Build list */
				while (low_char <= * pattern
					&& low_char != '\0')
				 /* avoid wrap-around beyond '\377' (loop!) */
					* expression ++ = low_char ++;
				pattern ++;
			}
			if (expression >= & exp_buffer [BLOCK_SIZE]) {
				reg_error (too_long);
				return FINE;
			}
		}
		pattern ++;			/* Skip ']' */
		/* Assign length of list in acct field */
		if ((* acct_field = (int) (expression - acct_field)) == 1) {
			reg_error ("Empty []");
			return FINE;
		}
		/* Assign negate and bracket field */
		* acct_field |= BRACKET;
		if (negate == True)
			* acct_field |= NEGATE;
		/*
		 * Add BRACKET to opcode of last char in field because
		 * a '*' may be following the list.
		 */
		previous (expression) |= BRACKET;
		break;
	default :
		* expression ++ = c + NORMAL;
	}
	if (expression == & exp_buffer [BLOCK_SIZE]) {
		reg_error (too_long);
		return FINE;
	}
  }
  /* NOTREACHED */
}

/*
 * Match gets as argument the program, pointer to place in current line to
 * start from and the method to search for (either FORWARD or REVERSE).
 * Match () will look through the whole file until a match is found.
 * NIL_LINE is returned if no match could be found.
 */
LINE *
match (program, string, method)
  REGEX * program;
  character * string;
  register FLAG method;
{
  register LINE * line = cur_line;
  character old_charbyte;		/* For saving chars */
  char * next = string;

/* Corrupted program */
  if (program->status == REG_ERROR)
	return NIL_LINE;

/* Check part of text first */
  if (! (program->status & BEGIN_LINE)) {
	if (method == FORWARD) {
		advance_char (& next);
		if (line_check (program, next, method) == MATCH)
			return cur_line;	/* Match found */
	}
	else if (! (program->status & END_LINE)) {
		old_charbyte = * string;	/* Save char and */
		* string = '\n';	/* Assign '\n' for line_check */
		if (line_check (program, line->text, method) == MATCH) {
			* string = old_charbyte; /* Restore char */
			return cur_line;    /* Found match */
		}
		* string = old_charbyte;    /* No match, but restore char */
	}
  }

/* No match in last (or first) part of line. Check out rest of file */
  do {
	line = (method == FORWARD) ? line->next : line->prev;
	if (line->text == NIL_PTR) {	/* Header/tail */
		status_msg ("Search wrapped around end of file");
		wrapped_search = True;
		continue;
		}
	if (line_check (program, line->text, method) == MATCH)
		return line;
  } while (line != cur_line && quit == False);

/* No match found. */
  return NIL_LINE;
}

/*
 * Line_check () checks the line (or rather string) for a match. Method
 * indicates FORWARD or REVERSE search. It scans through the whole string
 * until a match is found, or the end of the string is reached.
 */
int
line_check (program, string, method)
  register REGEX * program;
  char * string;
  FLAG method;
{
  char * textp = string;

/* Assign start_ptr field. We might find a match right away! */
  program->start_ptr = textp;

/* If the match must be anchored, just check the string. */
  if (program->status & BEGIN_LINE)
	return check_string (program, string, NIL_INT);

  if (method == REVERSE) {
	/* First move to the end of the string */
	for (textp = string; * textp != '\n'; textp ++)
		;
	/* Start checking string until the begin of the string is met */
	while (textp >= string) {
		program->start_ptr = textp;
		if (check_string (program, textp, NIL_INT))
			return MATCH;
		if (textp == string)
			textp --;	/* just flag exit */
		else
			precede_char (& textp, string);
	}
  }
  else {
	/* Move through the string until the end of it is found */
	while (quit == False && * textp != '\0') {
		program->start_ptr = textp;
		if (check_string (program, textp, NIL_INT))
			return MATCH;
		if (* textp == '\n')
			break;
		advance_char (& textp);
	}
  }

  return NO_MATCH;
}

/*
 * Check_string () checks if a match can be found in the given string.
 * Whenever a STAR is found during matching, then the begin position of
 * the string is marked and the maximum number of matches is performed.
 * Then the function star () is called which starts to finish the match
 * from this position of the string (and expression).
 * Check () returns MATCH for a match, NO_MATCH if the string
 * couldn't be matched or REG_ERROR for an illegal opcode in expression.
 */
int
check_string (program, string, expression)
  REGEX * program;
  character * string;
  int * expression;
{
  register int opcode;		/* Holds opcode of next expr. atom */
  character c;			/* Char that must be matched */
  character * mark;		/* For marking position */
  int star_fl;			/* A star has been born */

  if (expression == NIL_INT)
	expression = program->result.expression;

/* Loop until end of string or end of expression */
  while (quit == False && ! (* expression & DONE) &&
				   * string != '\0' && * string != '\n') {
	c = * expression & LOW_BYTE;	  /* Extract match char */
	opcode = * expression & HIGH_BYTE; /* Extract opcode */
	if ((star_fl = (opcode & STAR)) != 0) {  /* Check star occurrence */
		opcode &= ~STAR;	  /* Strip opcode */
		mark = string;		  /* Mark current position */
	}
	expression ++;		/* Increment expr. */
	switch (opcode) {
	case NORMAL :
		if (star_fl)
			while (* string == c) {	/* Skip all matches */
/*				advance_char (& string);	*/
				string ++;
			}
		else {
			if (* string != c)
				return NO_MATCH;
/*			advance_char (& string);	*/
			string ++;
		}
		break;
	case DOT :
		advance_char (& string);
		if (star_fl)			/* Skip to eoln */
			while (* string != '\0' && * string != '\n')
				advance_char (& string);
		break;
	case NEGATE | BRACKET :
	case BRACKET :
	     if (star_fl)
		while (in_exprlist (expression, string, c, opcode) == MATCH)
			advance_char (& string);
	     else
		if (in_exprlist (expression, string, c, opcode) == NO_MATCH)
			return NO_MATCH;
		advance_char (& string);
		expression += c - 1;	/* Add length of range list */
		break;
	default :
		/* panic ("Corrupted search program", NIL_PTR); */
		ring_bell ();
		error ("Corrupted search program", NIL_PTR);
		sleep (2);
		return NO_MATCH;
	}
	if (star_fl)
		return star (program, mark, string, expression);
  }
  if (* expression & DONE) {
	program->end_ptr = (char *) string;	/* Match ends here */
	/*
	 * We might have found a match. The last thing to do is check
	 * whether a '$' was given at the end of the expression, or
	 * the match was found on a null string. (E.g. [a-z]* always
	 * matches) unless a ^ or $ was included in the pattern.
	 */
	if ((* expression & EOLN) && * string != '\n' && * string != '\0')
		return NO_MATCH;
	if (string == (character *) program->start_ptr
					&& ! (program->status & BEGIN_LINE)
					&& ! (* expression & EOLN))
		return NO_MATCH;
	return MATCH;
  }
  return NO_MATCH;
}

/*
 * Star () calls check_string () to find out the longest match possible.
 * It searches backwards until the (in check_string ()) marked position
 * is reached, or a match is found.
 */
int
star (program, end_position, string, expression)
  REGEX * program;
  register char * end_position;
  char * string;
  int * expression;
{
  do {
	if (check_string (program, string, expression))
		return MATCH;
	precede_char (& string, end_position);
  } while (string != end_position);

  return NO_MATCH;
}

/*
 * In_list () checks if the given character is in the list of []. If it is
 * it returns MATCH. If it isn't it returns NO_MATCH. These returns values
 * are reversed when the NEGATE field in the opcode is present.
 *
 * multi-byte characters are stored in separate successive character fields
 * of the compiled search expression list
 */
int
in_exprlist (exprlist, s, exprlist_length, opcode)
  register int * exprlist;
  character * s;
  register int exprlist_length;
  int opcode;
{
  character * spoi;
  FLAG charmatch;

  if (* s == '\0' || * s == '\n')	/* End of string, never matches */
	return NO_MATCH;

  if (utf8_text == True)
	while (exprlist_length > 1) {	/* > 1, don't check acct_field */
		charmatch = True;
		if ((* exprlist & LOW_BYTE) != * s)
			charmatch = False;
		if ((* exprlist & 0xC0) == 0xC0) { /* multi-byte char */
			spoi = s;
			exprlist ++; exprlist_length --;
			spoi ++;
			while ((* exprlist & 0xC0) == 0x80) {
				/* UTF-8 sequence byte */
				if (charmatch == True) {
					if ((* exprlist & LOW_BYTE) != * spoi)
						charmatch = False;
					spoi ++;
				}
				exprlist ++; exprlist_length --;
			}
		} else {
			exprlist ++; exprlist_length --;
		}
		if (charmatch == True)
			return (opcode & NEGATE) ? NO_MATCH : MATCH;
	}
  else if (Chinese == True)
	while (exprlist_length > 1) {	/* > 1, don't check acct_field */
		if ((* exprlist & LOW_BYTE) == * s)
		    if ((* exprlist & 0x80) == 0x80) { /* 2 byte char */
			exprlist ++; exprlist_length --;
			spoi = s; spoi ++;
			if ((* exprlist & LOW_BYTE) == * s)
				return (opcode & NEGATE) ? NO_MATCH : MATCH;
		    } else
			return (opcode & NEGATE) ? NO_MATCH : MATCH;
		exprlist ++;
		exprlist_length --;
	}
  else
	while (exprlist_length > 1) {	/* > 1, don't check acct_field */
		if ((* exprlist & LOW_BYTE) == * s)
			return (opcode & NEGATE) ? NO_MATCH : MATCH;
		exprlist ++;
		exprlist_length --;
	}

  return (opcode & NEGATE) ? MATCH : NO_MATCH;
}


/*  ==================================================================	*
 *				End					*
 *  ==================================================================	*/
